# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import minimize
import math

plt.ion()

# Support Vector Machine Model
class model:
	# kernal, e.g. (X' . X) for linear SVM
	def __init__(self, kernal):
		self.kernal = kernal

	# svm equation to be minimized
	# 1 / 2 * (W'  W) - (summation of lambdas)
	# where W = summation of lambda*y*x
	# lambdas are lagrange multipliers
	def svm_equation(self, lambdas, x_train, y_train):
		ret = 0
		N = len(x_train)
		for i in range(N):
			for j in range(N):
				ret += lambdas[i] * lambdas[j] * y_train[i] * y_train[j] * self.kernal(x_train[i], x_train[j])
		ret /= 2
		for i in range(N):
			ret -= lambdas[i]
		return ret
	
	# constraint --> summation of lambda*y = 0
	def svm_constraint(self, lambdas, y_train):
		total = 0
		for l, y in zip(lambdas[:-1], y_train):
			total += l * y
		return total

	def fit(self, x_train, y_train):
		# bounds fro lambda values
		bounds = np.array([(0, 100000)] * (len(x_train) + 1))
		cons = ({'type':'eq', 'fun':self.svm_constraint, 'args':[y_train]})
		# one more lambda for mu (for constaint)
		self.lambdas = minimize(self.svm_equation, np.random.rand(len(x_train) + 1), args=(x_train, y_train), bounds=bounds, constraints=cons).x
		# list of support vectors
		self.support_vectors = np.empty((0, len(x_train[0])), int)
		# list of (lambda * y)
		self.multipliers = np.empty((0), int)
		# save support vectors and multipliers where lambda != 0
		for i, l in enumerate(self.lambdas[:-1]):
			if not l or l < 1e-5: continue
			self.support_vectors = np.append(self.support_vectors, [x_train[i]], axis=0)
			self.multipliers = np.append(self.multipliers, [l * y_train[i]], axis=0)
		# Y = W' . X + B
		# B is bias
		self.bias = 0
		for i, l in enumerate(self.lambdas[:-1]):
			if not l or l < 1e-5: continue
			self.bias = y_train[i] - np.sum(np.array([self.multipliers]).T * self.kernal(self.support_vectors, np.array([x_train[i]])))
			break

	# y_pred = signum(W' . X + B)
	def predict(self, x_test):
		out = self.bias + np.sum(np.array([self.multipliers] * len(x_test)).T * self.kernal(self.support_vectors, x_test), axis=0)
		return np.array([1 if ele >= 0 else -1 for ele in out])

	# find accuracy over test dataset
	def evaluate(self, x_test, y_test):
		out = self.bias + np.sum(np.array([self.multipliers] * len(x_test)).T * self.kernal(self.support_vectors, x_test), axis=0)
		return np.count_nonzero(np.array([1 if ele >= 0 else -1 for ele in out]) == y_test) / len(x_test) * 100

	def plot_data(self, x, y, type):
		y_pred = self.predict(x)
		done_p = False
		done_n = False
		for i in range(len(x)):
			if y[i] == 1:
				plt.scatter(x[i][0], x[i][1], color='blue' if type == 'train' else 'red', marker='x', label=type+"ing data, y=+1" if not done_p else "")
				done_p = True
			else:
				plt.scatter(x[i][0], x[i][1], color='blue' if type == 'train' else 'red', marker='+', label=type+"ing data, y=-1" if not done_n else "")
				done_n = True
		plt.legend()
		plt.pause(1e-10)

	def plot_decision_boundary(self):
		delta = 0.025
		x1, x2 = np.meshgrid(np.arange(-0.1, 1.1, delta), np.arange(-0.1, 1.1, delta))
		X = np.empty((0, 2))
		for i in range(x1.shape[0]):
			for j in range(x1.shape[1]):
				X = np.append(X, np.array([[x1[i][j], x2[i][j]]]), axis=0)
		out = self.bias + np.sum(np.array([self.multipliers] * len(X)).T * self.kernal(self.support_vectors, X), axis=0)
		out2 = np.empty((0, x1.shape[1]))
		ptr = 0
		for i in range(x1.shape[0]):
			row = np.array([])
			for j in range(x1.shape[1]):
				row = np.append(row, out[ptr])
				ptr += 1
			out2 = np.append(out2, [row], axis=0)
		cont = plt.contour(x1, x2, out2, [0])
		cont.collections[0].set_label('Predicted Boundary')
		cont = plt.contourf(x1, x2, out2, [-1, 0, 1], alpha=0.3)
		plt.legend()
		plt.pause(1e-10)


# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

from numpy import dot
import numpy as np

# gradient at last (output) node, for sigmoid (use accordingly)
def grad(Y, y):
	return (Y - y) * y * (1 - y)

# softmax for output layer e^x / [summation of e^x]
def softmax(X):
	base = np.sum(np.exp(X))
	return np.exp(X) / base

# multiple activation functions
def activate(X, activation_func="sigmoid"):
	if activation_func == "sigmoid":
		return 1 / (1 + np.exp(-X))
	elif activation_func == "tanh":
		return np.tanh(X)
	elif activation_func == "sin":
		return np.sin(X)
	elif activation_func == "cos":
		return np.cos(X)

# initialize random weights
def random_weights(neurons_per_layer, bias):
	W = []
	for layer in range(1, len(neurons_per_layer)):
		W.append(np.random.randn(neurons_per_layer[layer], neurons_per_layer[layer - 1] + (1 if bias else 0)))
	return W

# Neural Network Model
class model:
	def __init__(self, number_of_layers, alpha, neurons_per_layer, activations_per_layer, bias=True, W=[]):
		# total number of layer of NN, excluding input layer and including hidden and output layer
		self.number_of_layers = number_of_layers
		# learning rate
		self.alpha = alpha
		# boolean (to use bias at each layer or not)
		self.bias = bias
		# list of number of neaurons at each layer
		self.neurons_per_layer = neurons_per_layer.copy()
		# activation functions to be used at each layer
		self.activation_func = activations_per_layer.copy()
		# weights, if already trained
		if W: self.W = W
		# weights, if randomly needed
		else: self.W = random_weights(self.neurons_per_layer, self.bias)

	# forward propagation
	def forward(self, X):
		# actiavte(z)
		self.a = []
		# (W' . X)
		self.z = []
		# delta at each layer
		self.delta = []
		# initial input
		self.a.append(X)
		for layer in range(self.number_of_layers):
			if self.bias:
				self.a[layer] = np.append(self.a[layer], np.ones((1, 1)), axis=0)
			self.z.append(dot(self.W[layer], self.a[layer]))
			# print(activate(self.z[layer], self.activation_func[layer]).shape)
			self.a.append(activate(self.z[layer], self.activation_func[layer]))
		if self.activation_func[-1] == "sigmoid":
			return self.a[-1]
		else:
			# for tanh
			return (self.a[-1] + 1) / 2

	# backward propagation
	def backward(self, Y, y):
		# find gradient at last (output) layer
		gradient = grad(Y, y)
		self.delta = [np.array([]) for layer in range(self.number_of_layers + 1)]
		# find delta at each layer and then update weights
		if self.bias:
			self.delta[self.number_of_layers] = np.append(gradient, np.ones((1, 1)), axis=0)
			for layer in range(self.number_of_layers - 1, -1, -1):
				if self.activation_func[layer] == "sigmoid":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * self.a[layer] * (1 - self.a[layer])
				elif self.activation_func[layer] == "tanh":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * (1- self.a[layer] * self.a[layer])
				elif self.activation_func[layer] == "sin":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * np.cos(self.a[layer])
				elif self.activation_func[layer] == "cos":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * (-np.sin(self.a[layer]))
			for layer in range(self.number_of_layers):
				self.W[layer] = self.W[layer] + self.alpha * dot(self.delta[layer + 1][:-1], np.transpose(self.a[layer]))
		else:
			self.delta[self.number_of_layers] = gradient
			for layer in range(self.number_of_layers - 1, -1, -1):
				if self.activation_func[layer] == "sigmoid":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * self.a[layer] * (1 - self.a[layer])
				elif self.activation_func[layer] == "tanh":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * (1- self.a[layer] * self.a[layer])
				elif self.activation_func[layer] == "sin":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * np.cos(self.a[layer])
				elif self.activation_func[layer] == "cos":
					self.delta[layer] = dot(np.transpose(self.W[layer]), self.delta[layer + 1][:-1]) * (-np.sin(self.a[layer]))
			for layer in range(self.number_of_layers):
				self.W[layer] = self.W[layer] + self.alpha * dot(self.delta[layer + 1], np.transpose(self.a[layer]))



# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

import numpy as np

class model:
	def __init__(self, lr):
		self.lr = lr
		self.W = np.random.rand(3)

	# update weight if Y * (W' . X) < 0
	def fit(self, X, Y):
		X = np.array([np.append(x, 1) for x in X])
		done = False
		while not done:
			done = True
			for i, x in enumerate(X):
				if (np.dot(self.W, X[i]) >= 0) != (Y[i] == 1):
					self.W += np.dot(Y[i], X[i])
					done = False
					break
		return self.W

	# y_pred = signum(W' . x_test)
	def predict(self, x_test):
		x_test = np.array([np.append(x, 1) for x in x_test])
		y_pred = []
		for i, x in enumerate(x_test):
			y_pred.append(1 if np.dot(self.W, x_test[i]) >= 0 else -1)
		return np.array(y_pred)

	# find accuracy over test dataset
	def evaluate(self, x_test, y_test):
		y_pred = self.predict(x_test)
		return np.count_nonzero(y_pred == y_test) / len(y_pred) * 100


# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

import numpy as np
from vin_ml.utils import sigmoid, mse
from scipy.optimize import fmin_cg

# -1 / N * [summation of log of sigmoid(W' . X)]
def logisitic_loss(W, X, Y):
	if (W.ndim == 1): W = W.reshape((len(W), 1))
	N = len(X)
	J = -1 / N * np.sum(np.log(sigmoid((Y * np.dot(X, W)))))
	return J.item(0)

# 1 / N * (summation of (-Y . X . sigmoid(-Y . W' . X)))
def logistic_gradient(W, X, Y):
	W = W.reshape((len(W), 1))
	N = len(X)
	delta = 1 / N * np.dot(-(Y * X).T, sigmoid(-Y * np.dot(X, W)))
	return np.ndarray.flatten(delta)

# logistic regression model
class model:
	def fit(self, X, Y):
		# logistic loss optimizer (gradient descent)
		self.W = fmin_cg(f = logisitic_loss, x0 = np.zeros(len(X[0])), fprime = logistic_gradient, args = (X, Y)).reshape((len(X[0]), 1))
		return self.W

	# threshold over sigmoid(W' . X)
	def predict(self, X):
		y_pred = sigmoid(np.dot(X, self.W))
		y_pred = np.array([[1 if ele >= 0.5 else -1 for ele in y_pred]]).T
		return y_pred

	# find accuracy over test dataset
	def evaluate(self, X, Y):
		y_pred = self.predict(X)
		return np.count_nonzero(Y == y_pred) / len(X) * 100

	# find mean square loss over test dataset
	def loss(self, X, Y):
		y_pred = self.predict(X)
		return mse(y_pred, Y)


# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

# KERNALS used in svm demo
import numpy as np

def linear(x, y):
	return np.dot(x, y.T)

def polynomial(x, y, p = 2):
	return (1 + np.dot(x, y.T)) ** p

# add more kernals as per usage


# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

from math import log2
import numpy as np
import pydotplus as pdp
from math import inf

NODE_FILL_COLOR = ["red", "green", "blue", "green", "grey"]
FONT_COLOR = ["white", "blue", "white", "blue", "white"]

def most_common(L):
  return max(set(L), key=list(L).count)

# this is recurive implementation of decision tree, with current root node as instance of this model, childrens are also instance of this model
class model:
	def __init__(self, data, feature_index, target, levels, features, most_frequent_target_value):
		# childrens of this node after classifying data by maximum IGR
		self.children = {}
		# entropy at this node
		self.entropy = 0
		self.is_leaf = False
		# Target feature value
		self.target_value = ""
		# Maximum IGR feature name
		self.max_igr_feature = ""
		# Target feature name
		self.target = target
		# data contains all the data to be trained on decision tree
		# a 2D array, with a row containing single instance (values) of data
		self.data = data.copy()
		# feature index, dictionary with key as feature name and value as index column of that feature in 2D data (0-based)
		self.feature_index = feature_index.copy()
		# dictionary with key as feature name and value as a list of values which that feature can take
		self.levels = levels.copy()
		# list of all features
		self.features = features.copy()
		self.most_frequent_target_value = most_frequent_target_value
		if not len(data):
			self.is_leaf = True
			self.target_value = most_frequent_target_value
			# print("No Data, Set Target Value to", self.target_value)
			return
		if not len(features):
			self.is_leaf = True
			self.target_value = most_common(data[:, -1].flatten())
			# print("No Data, Set Target Value to", self.target_value)
			return
		# find entropy of whole dataset at this node
		if len(self.data):
			self.entropy = self.find_entropy_whole_dataset(self.data, self.target)
		if (self.entropy == 0):
			# if node has entropy 0 w.r.t. target, then this is leaf node
			self.is_leaf = True
			self.target_value = self.data[0][feature_index[target]]
			# print("Set Target Value to", self.target_value)
		# else:
			# print("H (", self.target, ") =", self.entropy)

	# entroy of data w.r.t. some feature
	def find_entropy_whole_dataset(self, data, feature):
		# negative summation of log probablities multiplied with probability
		count = {}
		for level in self.levels[feature]:
			count[level] = 0
		data_length = len(data)
		entropy = 0
		for row in data:
			count[row[self.feature_index[feature]]] += 1
		for level in self.levels[feature]:
			probability = count[level] / data_length
			if not probability:
				continue
			entropy += -1 * probability * log2(probability)
		return entropy

	# filter data with the feature value, take only that subset of data which has specified feature value
	def filter_data(self, feature, value):
		filtered_data = []
		for row in self.data:
			if row[self.feature_index[feature]] == value:
				filtered_data.append(row)
		return filtered_data

	# calculates remaining entropy of feature2 after dividing data with feature1 
	def find_entropy_wrt_feature(self, data, feature1, feature2):
		count = {}
		for level in self.levels[feature1]:
			count[level] = 0
		for row in data:
			count[row[self.feature_index[feature1]]] += 1 
		data_length = len(data)
		entropy = 0
		for level in self.levels[feature1]:
			probability = count[level] / data_length
			filtered_data = self.filter_data(feature1, level)
			if not probability:
				continue
			entropy += probability * self.find_entropy_whole_dataset(filtered_data, feature2)
		return entropy

	def find_information_gain(self, feature1, feature2):
		return self.entropy - self.find_entropy_wrt_feature(self.data, feature1, feature2)

	def find_information_gain_ratio(self, feature):
		entropy_wrt_feature = self.find_entropy_whole_dataset(self.data, feature)
		return self.find_information_gain(feature, self.target) / entropy_wrt_feature if entropy_wrt_feature != 0 else inf

	# build tree
	def build(self):
		if self.is_leaf:
			return self
		information_gain_ratio = {}
		# find igr w.r.t. each feature
		for feature in self.features:
			information_gain_ratio[feature] = self.find_information_gain_ratio(feature)
			# print("IG (", feature, ") =", information_gain_ratio[feature])
		#select feature with maximum igr
		max_igr, self.max_igr_feature = max(zip(information_gain_ratio.values(), information_gain_ratio.keys()))
		# print("Best Feature: ", self.max_igr_feature)
		max_igr_feature_index = self.feature_index[self.max_igr_feature]
		# divide data and its properties w.r.t. each level of maximum igr feature and recursively build tree
		for level in self.levels[self.max_igr_feature]:
			if level not in np.array(self.data)[:, self.feature_index[self.max_igr_feature]]:
				continue
			# print("On Branch", self.max_igr_feature, "=", level)
			filtered_data = np.array(self.filter_data(self.max_igr_feature, level))
			if filtered_data.size:
				filtered_data = np.delete(filtered_data, [max_igr_feature_index, max_igr_feature_index], axis = 1)
			features = np.delete(np.array(self.features), [max_igr_feature_index, max_igr_feature_index])
			feature_index = {feature:index for feature, index in self.feature_index.items() if index != max_igr_feature_index}
			feature_index = {feature:(index if index < max_igr_feature_index else index - 1) for feature, index in feature_index.items()}
			self.children[level] = model(filtered_data, feature_index, self.target, self.levels, features, self.most_frequent_target_value).build()
		return self

	# dfs over tree to predict query
	def predict(self, query):
		if self.is_leaf:
			return self.target_value
		return self.children[query[self.max_igr_feature]].predict(query)

	# create real nodes into 'graph' variable using dfs
	def plot(self, graph, parent=None, parent_label=None, parent_child_edge=None, query=None, root=None):
		if self.is_leaf:
			if query and query[parent] == parent_child_edge:
				if query and self.target_value == query[self.target]:
					graph.add_node(pdp.Node(parent_label + parent_child_edge + self.target_value, label=self.target_value, shape="box", style="filled", fillcolor="skyblue", fontcolor="black"))
				else:
					graph.add_node(pdp.Node(parent_label + parent_child_edge + self.target_value, label=self.target_value, shape="box", style="filled", fillcolor="pink", fontcolor="black"))
				graph.add_edge(pdp.Edge(parent_label, parent_label + parent_child_edge + self.target_value, label=parent_child_edge, fontsize="14", color="black", fontcolor="purple"))
			else:
				graph.add_node(pdp.Node(parent_label + parent_child_edge + self.target_value, label=self.target_value, shape="box", style="filled", fillcolor="yellow", fontcolor="black"))
				graph.add_edge(pdp.Edge(parent_label, parent_label + parent_child_edge + self.target_value, label=parent_child_edge, fontsize="14", color="brown", fontcolor="purple"))
		else:
			index = root.feature_index[self.max_igr_feature]%len(NODE_FILL_COLOR)
			if parent_label:
				if query and query[parent] == parent_child_edge:
					graph.add_node(pdp.Node(parent_label + parent_child_edge + self.max_igr_feature, label=self.max_igr_feature, style="filled", fillcolor="black", fontcolor="white", color="white"))
					graph.add_edge(pdp.Edge(parent_label, parent_label + parent_child_edge + self.max_igr_feature, label=parent_child_edge, fontsize="14", color="black", fontcolor="purple"))
				else:
					graph.add_node(pdp.Node(parent_label + parent_child_edge + self.max_igr_feature, label=self.max_igr_feature, style="filled", fillcolor=NODE_FILL_COLOR[index], fontcolor=FONT_COLOR[index], color="white"))
					graph.add_edge(pdp.Edge(parent_label, parent_label + parent_child_edge + self.max_igr_feature, label=parent_child_edge, fontsize="14", color="brown", fontcolor="purple"))
				for level in self.levels[self.max_igr_feature]:
					if level not in self.children:
						continue
					if query and level == query[self.max_igr_feature]:
						self.children[level].plot(graph, parent=self.max_igr_feature, parent_label=parent_label + parent_child_edge + self.max_igr_feature, parent_child_edge=level, query=query, root=root)
					else:
						self.children[level].plot(graph, parent=self.max_igr_feature, parent_label=parent_label + parent_child_edge + self.max_igr_feature, parent_child_edge=level, root=root)
			else:
				if query:
					graph.add_node(pdp.Node(self.max_igr_feature, label=self.max_igr_feature, style="filled", fillcolor="black", fontcolor="white", color="white"))
				else:
					graph.add_node(pdp.Node(self.max_igr_feature, label=self.max_igr_feature, style="filled", fillcolor=NODE_FILL_COLOR[index], fontcolor=FONT_COLOR[index], color="white"))
				for level in self.levels[self.max_igr_feature]:
					if level not in self.children:
						continue
					if query and level == query[self.max_igr_feature]:
						self.children[level].plot(graph, parent=self.max_igr_feature, parent_label=self.max_igr_feature, parent_child_edge=level, query=query, root=root)
					else:
						self.children[level].plot(graph, parent=self.max_igr_feature, parent_label=self.max_igr_feature, parent_child_edge=level, root=root)




# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

# utility functions
import numpy as np

def mse(X1, X2):
	diff = (X1 - X2)
	E = 1 / len(X1) * np.dot(diff.T, diff)
	return E.item(0)

def sigmoid(z):
	return (1 / (1 + np.exp(-z)))


# Author: Vinayak Sachdeva
# Btech, MNIT Jaipur

import numpy as np
from numpy.linalg import inv
from vin_ml.utils import mse

# linear regression model
class model:
	def __init__(self, regularization=0):
		self.regularization = regularization

	# inv(X' . X + lambda . I) . X' . Y
	def fit(self, X, Y):
		self.W = np.dot(np.dot(inv(np.dot(X.T, X) + self.regularization * np.eye(X.shape[1])), X.T), Y)
		return self.W

	# (W' . X)
	def predict(self, x_test):
		return np.dot(x_test, self.W)

	# 1 / len(x_test) * (y_pred - y_test) * (y_pred - y_test)
	def loss(self, X, Y):
		y_pred = self.predict(X)
		return mse(y_pred, Y)


import setuptools
setuptools.setup(
     name='vin_ml',  
     version='0.1',
     author="Vinayak Sachdeva",
     author_email="vishusachdeva228@gmail.com",
     description="My ML Lab Codes",
     url='https://github.com/Vinayak-sophos/vin_ml',
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "Operating System :: OS Independent",
     ],
 )